import re
from typing import List, Optional
from ipaddress import IPv4Address

from netaddr import EUI, mac_unix_expanded

from .base import BaseDevice, ARPTableEntry, MACTableEntry, Interface


VALID_PORTS = re.compile(r"^(ge|xe|ae|et)")

ADMIN_STATUS = {
    "up": "enabled",
    "down": "disabled",
}

OPER_STATUS = {"up": "up", "down": "down"}

DUPLEX = {
    "Half-duplex": "half",
    "Full-duplex": "full",
}

SPEED = {
    "Auto": "auto",
    "1000mbps": "1000",
    "100mbps": "100",
    "10mbps": "10",
}


class Junos(BaseDevice):
    def get_interface_details(self, interface: Optional[str] = None) -> List[Interface]:
        """
        issue the junos RPC <get-interface-information> and return normalized output

        physical-interface {
            name ge-0/0/1
            admin-status up
            oper-status down
            description DEFAULT
            link-mode Half-duplex
            speed Auto
            ...
        """
        output = []

        rpc = (
            self.device.rpc.jrpc__rpc_get_interface_information.get_interface_information
        )
        inp = rpc.get_input()

        # First we're getting interface information
        if interface:
            inp.interface_name = interface

        self.log.debug(f"sending <get-inteface-information> to {self.device.name}")
        reply = rpc(inp)

        # save the fields we care about in our output as normalized data
        for i in reply.interface_information.physical_interface:
            if not (VALID_PORTS.match(i.name)):
                continue

            if (not (interface)) or (interface and interface == i.name):
                record = Interface(
                    name=i.name,
                    description=i.description,
                    admin_status=ADMIN_STATUS[i.admin_status],
                    oper_status=OPER_STATUS[i.oper_status],
                    speed=SPEED.get(i.speed, "auto"),
                    duplex=DUPLEX.get(i.link_mode, "auto"),
                )
                output.append(record)

        return output

    def get_mac_table(self, mac: Optional[str] = None) -> List[MACTableEntry]:
        """
        issue the junos RPC <get-ethernet-switching-table-information><brief/> and return normalized output

        optionally filter the by the given mac address

        ...
        l2ng-mac-entry {
            l2ng-l2-mac-vlan-name VLAN0009
            l2ng-l2-mac-address cc:88:c7:cc:15:b3
            l2ng-l2-mac-flags D
            l2ng-l2-mac-age -
            l2ng-l2-mac-logical-interface ge-0/0/37.0
            l2ng-l2-mac-fwd-next-hop 0
            l2ng-l2-mac-rtr-id 0
        }
        ...
        """

        rpc = (
            self.device.rpc.jrpc__rpc_get_ethernet_switching_table_information.get_ethernet_switching_table_information
        )
        inp = rpc.get_input()
        inp.brief.create()

        if mac is not None:
            normalized_mac = EUI(mac, dialect=mac_unix_expanded)
            inp.address = str(normalized_mac)

        self.log.debug(
            f"sending <get-ethernet-switching-table-information> to {self.device.name}"
        )
        reply = rpc(inp)

        mac_table = reply.l2ng_l2ald_rtb_macdb.l2ng_l2ald_mac_entry_vlan[
            0
        ].l2ng_mac_entry

        output = []
        if not (mac_table):
            return output

        for entry in mac_table:
            entry_interface = entry.l2ng_l2_mac_logical_interface.replace(".0", "")
            if not (VALID_PORTS.match(entry_interface)):
                continue

            vid = self.device.config.configuration.vlans.vlan[
                entry.l2ng_l2_mac_vlan_name
            ].vlan_id

            entry_mac = entry.l2ng_l2_mac_address
            normalized_entry_mac = EUI(entry_mac, dialect=mac_unix_expanded)

            if mac is None or normalized_mac == normalized_entry_mac:
                record = MACTableEntry(
                    address=normalized_entry_mac,
                    vlan_id=vid,
                    interface=entry_interface,
                )
                output.append(record)

        return output

    def get_arp_table(
        self, address: Optional[str] = None, vrf: Optional[str] = None
    ) -> List[ARPTableEntry]:
        """
        issue the junos RPC <get-arp-table-information> and return normalized output

        ...
        arp-table-entry {
            mac-address 3c:8b:7f:24:41:67
            ip-address 10.233.0.2
            hostname v-mgmt-arbl.dl-arbl-1.umnet.umich.edu
            interface-name irb.3 [ge-0/2/0.0]
            arp-table-entry-flags {
                none
            }
        }
        ...
        """

        rpc = (
            self.device.rpc.jrpc__rpc_get_arp_table_information.get_arp_table_information
        )
        inp = rpc.get_input()
        if address:
            inp.hostname = address

        self.log.debug(f"sending <get-arp-table-information> to {self.device.name}")
        reply = rpc(inp)

        output = []
        for a in reply.arp_table_information.arp_table_entry:

            if (not (address)) or (address and address == a.ip_address):
                record = ARPTableEntry(
                    ip_address=IPv4Address(a.ip_address),
                    mac_address=EUI(a.mac_address, dialect=mac_unix_expanded),
                    interface=a.interface_name,
                    vrf=vrf,
                )
                output.append(record)

        return output


class Junos12(Junos):
    def get_mac_table(self, mac: Optional[str] = None) -> List[MACTableEntry]:
        """
        issue the junos RPC <get-interface-information> and return normalized output

        ...
        mac-table-entry {
            mac-vlan VLAN0020
            mac-address 00:50:f9:01:3e:94
            mac-type Learn
            mac-age 0
            mac-interfaces-list {
                mac-interfaces ge-0/0/38.0
            }
        }
        ...

        """

        if mac is not None:
            normalized_mac = EUI(mac, dialect=mac_unix_expanded)

        rpc = (
            self.device.rpc.jrpc__rpc_get_ethernet_switching_table_information.get_ethernet_switching_table_information
        )
        inp = rpc.get_input()
        reply = rpc(inp)

        mac_table = (
            reply.ethernet_switching_table_information.ethernet_switching_table.mac_table_entry
        )

        output = []
        for entry in mac_table:
            entry_interface = entry.mac_interfaces_list.mac_interfaces.replace(".0", "")
            if (entry.mac_type != "Learn") or not (VALID_PORTS.match(entry_interface)):
                continue

            normalized_entry_mac = EUI(entry.mac_address, dialect=mac_unix_expanded)
            vid = self.device.config.configuration.vlans.vlan[entry.mac_vlan].vlan_id

            if mac is None or normalized_mac == normalized_entry_mac:
                record = MACTableEntry(
                    address=normalized_entry_mac,
                    vlan_id=vid,
                    interface=entry_interface,
                )
                output.append(record)

        return output



import re
import logging
from typing import Any, Optional, List, Union
from dataclasses import dataclass
from ipaddress import IPv4Address

from netaddr import EUI
import ncs


@dataclass
class MACTableEntry:
    """
    Represents a normalized entry in any given devices' MAC address table
    """

    address: EUI  # EUI("cc:88:c7:cc:16:ac", dialect=mac_unix_expanded)
    interface: str  # "Gi1/0", "ge-0/0/18", etc
    vlan_id: Optional[int] = None  # range(1, 4096)


@dataclass
class ARPTableEntry:
    """
    Represents a normalized entry in any given devices' ARP table
    """

    ip_address: IPv4Address
    mac_address: EUI  # EUI("cc:88:c7:cc:16:ac", dialect=mac_unix_expanded)
    interface: str  # "Gi1/0", "ge-0/0/18", etc
    vrf: Optional[str] = None


@dataclass
class Interface:
    """
    Represents normalized operation data about a device interface
    """

    name: str  # inteface id, e.g. "ge-0/0/18" or "Gi1/0"
    admin_status: str  # "enabled" or "disabled"
    oper_status: str  # "up" or "down"
    duplex: str  # auto/half/full
    speed: str  # 10/100/1000/auto
    is_logical: bool = False  # software interfaces (Vlan,irb, etc) set to True
    description: Optional[str] = None  # e.g. "1401A: hallway printer"


class BaseDevice:
    def __init__(self, device: ncs.maagic.Node, log: logging.Logger):
        self.device = device
        self.log = log

    @staticmethod
    def normalize_mac(mac: str) -> str:
        mac = re.sub(r"[\.\-\:]", "", mac)
        return f"{mac[0:2]}:{mac[2:4]}:{mac[4:6]}:{mac[6:8]}:{mac[8:10]}:{mac[10:12]}"

    def _run_cmd(self, command: str) -> Any:
        raise NotImplementedError

    def get_mac_table(self, mac: Optional[str] = None) -> List[MACTableEntry]:
        raise NotImplementedError

    def get_arp_table(
        self, address: Optional[str] = None, vrf: Optional[str] = None
    ) -> List[ARPTableEntry]:
        raise NotImplementedError

    def get_interface_details(self, interface: Optional[str] = None) -> List[Interface]:
        raise NotImplementedError

    def get_bfd_neighbors(self, interface: Optional[str] = None) -> Any:
        raise NotImplementedError

    def get_ospf_neighbors(self, interface: Optional[str] = None) -> Any:
        raise NotImplementedError

    def get_lldp_neighbors(self, interface: Optional[str] = None) -> Any:
        raise NotImplementedError

    def get_transceiver_details(self, interface: Optional[str] = None) -> Any:
        raise NotImplementedError



from typing import List, Optional
from ipaddress import IPv4Address

from ntc_templates.parse import parse_output
from netaddr import EUI, mac_unix_expanded

from .base import BaseDevice, ARPTableEntry, MACTableEntry, Interface


# note that since we're using 'show int status',
# admin and oper status leverage the same output column, whose values can be:
#  - connected (admin up, oper up)
#  - notconnect (admin up, oper down)
#  - disabled (admin down, oper down)
#  - suspended (admin up, oper down)
#  - err-disabled (admin up, oper down)
ADMIN_STATUS = {
    "connected": "enabled",
    "notconnect": "enabled",
    "disabled": "disabled",
    "suspended": "enabled",
    "err-disabled": "enabled",
}
OPER_STATUS = {
    "connected": "up",
    "notconnect": "down",
    "disabled": "down",
    "suspended": "down",
    "err-disabled": "down",
}
DUPLEX = {
    "half": "half",
    "full": "full",
    "auto": "full",
    "a-half": "half",
    "a-full": "full",
}
SPEED = {
    "10": "10",
    "100": "100",
    "1000": "1000",
    "10G": "auto",
    "auto": "auto",
    "a-10": "10",
    "a-100": "100",
    "a-1000": "1000",
    "a-10G": "auto",
}


def _remove_prompt(results: str) -> str:
    """
    remove the trailing prompt from the results of a live_status show
    command.  This only appears to be an issue for the IOS NED. i.e:

    ...
    Gi2/0/48        notconnect   1       auto   auto 10/100/1000BaseTX
    Twe2/1/1        connected    1       full    10G SFP-10GBase-SR
    Twe2/1/2        notconnect   1       auto   auto unknown
    Ap2/0/1         connected    1     a-full a-1000 App-hosting port
    al-shallow-1#    <-- (rude)
    """
    lines = results.splitlines()
    try:
        del lines[-1]
    except IndexError:
        pass
    return "\n".join(lines)


class IOS(BaseDevice):
    def _run_cmd(self, command: str) -> str:
        """
        use NCS live-status exec to issue a raw show command towards a device.
        platform-specific models are expected to parse this output and return
        structured data.

        :param command: string CLI command, e.g. 'show interface status'
        :returns: raw string output from the device
        """
        show = self.device.live_status.__getitem__("exec").show
        inp = show.get_input()
        inp.args = [command]

        results = show.request(inp)

        return _remove_prompt(results.result)

    def get_interface_details(self, interface: Optional[str] = None) -> List[Interface]:
        """
        Gathers interface operational data from an IOS device by parsing 'show interfaces status'

        :interface: optionally get data only for a single interface

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_interfaces_status

        ---
        parsed_sample:
          - port: "Gi1/0/1"
            name: ""
            status: "notconnect"
            vlan: "1"
            duplex: "auto"
            speed: "auto"
            type: "10/100/1000BaseTX"
            fc_mode: ""
        """
        # ncs automatically adds the 'show' to the front of the cmd
        command = f"interfaces {interface} status" if interface else "interfaces status"
        self.log.debug(f"sending 'show {command}' to {self.device.name}")
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_ios",
            command="show interfaces status",
            data=reply,
        )

        output = []
        for entry in parsed:
            interface = Interface(
                name=entry["port"],
                description=entry["name"],
                admin_status=ADMIN_STATUS[entry["status"]],
                oper_status=OPER_STATUS[entry["status"]],
                duplex=DUPLEX[entry["duplex"]],
                speed=SPEED[entry["speed"]],
            )

            if interface.name.startswith("Vlan"):
                interface.is_logical = True

            output.append(interface)

        return output

    def get_arp_table(
        self, address: Optional[str] = None, vrf: Optional[str] = None
    ) -> List[ARPTableEntry]:
        """
        Gathers ARP data from an IOS device by parsing the output of the CLI
        command 'show ip arp [ address | vrf <VRF>]'

        :address: optionally filter device output by MAC or IP address
        :vrf: optional string name of VRF table to look in

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_ip_arp

        TODO: add support(recursion?) for vrf='all' since ios makes you run a separate
              lookup command for every vrf (so rude)

        ---
        parsed_sample:
          - protocol: "Internet"
            address: "172.16.233.229"
            age: "-"
            mac: "0000.0c59.f892"
            type: "ARPA"
            interface: "Ethernet0/0"
        """
        # ncs automatically adds the 'show' to the front of the cmd
        command = "ip arp"
        if vrf is not None:
            command = command + f" vrf {vrf}"
        if address is not None:
            command = command + f" address {address}"
            normalized_address = IPv4Address(address)

        self.log.debug(f"sending 'show {command}' to {self.device.name}")
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_ios",
            command="show ip arp",
            data=reply,
        )

        output = []
        for entry in parsed:
            if entry["mac"] == "Incomplete":
                continue

            normalized_entry_address = IPv4Address(entry["address"])

            if address is None or normalized_address == normalized_entry_address:
                arp = ARPTableEntry(
                    ip_address=IPv4Address(entry["address"]),
                    mac_address=EUI(entry["mac"], dialect=mac_unix_expanded),
                    interface=entry["interface"],
                    vrf=vrf,
                )
                output.append(arp)

        return output

    def get_mac_table(self, mac: Optional[str] = None) -> List[MACTableEntry]:
        """
        Gathers dynamically-learned MAC address data from an IOS device by
        parsing the output of the CLI command 'show mac address-table dynamic
        [address <address>]'

        :mac: optionally filter device output by MAC address

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_mac-address-table

        ---
        parsed_sample:
        - destination_address: "30a3.30a3.a1c3"
            type: "dynamic"
            vlan: "666"
            destination_port:
            - "Te1/30"
        """
        # ncs automatically adds the 'show' to the front of the cmd
        command = "mac address-table dynamic"
        if mac is not None:
            normalized_mac = EUI(mac, dialect=mac_unix_expanded)
            command = command + f" address {mac}"

        self.log.debug(f"sending 'show {command}' to {self.device.name}")
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_ios",
            command="show mac-address-table",
            data=reply,
        )

        output = []
        for entry in parsed:
            try:
                vid = int(entry["vlan"])
            except ValueError:
                vid = None

            normalized_entry_mac = EUI(
                entry["destination_address"], dialect=mac_unix_expanded
            )
            # unless there's an on-going broadcast storm we should only be
            # learning an address on a single port -- not sure why the template
            # is modelled this way?
            interface = entry["destination_port"][0]

            if mac is None or normalized_mac == normalized_entry_mac:
                mac_entry = MACTableEntry(
                    address=normalized_entry_mac,
                    vlan_id=vid,
                    interface=interface,
                )
                output.append(mac_entry)

        return output



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['umnet_pyncs', 'umnet_pyncs.state', 'umnet_pyncs.state.models']

package_data = \
{'': ['*']}

install_requires = \
['multiprocessing-logging>=0.3.3,<0.4.0',
 'netaddr>=0.8.0,<0.9.0',
 'ntc-templates>=3.0.0,<4.0.0']

setup_kwargs = {
    'name': 'umnet-pyncs',
    'version': '0.1.11',
    'description': 'custom python module for NCS helpers',
    'long_description': '## Overview\n\nThis module is intended to be installed on the production NCS nodes and imported in other services/actions that need to gather state from the network.  It uses the NCS device manager and the standard python `multiprocessing` library to connect to devices in-parallel and issue commands, returning results as structured data.\n\n## Usage information\n\nBasic usage example in an NCS callback:\n\n``` python\nfrom umnet_ncs.state import StateManager\n...\n\n\nclass DemoAction(Action):\n    @Action.action\n    def cb_action(self, uinfo, name, kp, input, output, trans):\n        ...\n        with StateManager() as m:\n            interfaces = m.get_state(al_devices, ["get-interface-details"])\n            arp = m.get_state(dl_devices, ["get-arp-table"])\n            ...\n```\n\n## Supported commands\n\nCurrently supported commands are:\n- `get-mac-table`\n- `get-arp-table`\n- `get-interface-details`\n- `get-transciever-details`\n- `get-lldp-neighbors`\n- `get-bfd-neighbors`\n- `get-ospf-neighbors`\n\nFor any given command, the various platform-specific models are responsible for implementing how the data is fetched and parsed from the remote device.  Each command corresponds to a method that can be invoked to retrieve the data, e.g. `get-interface-details` maps to the `get_interface_details()` instance method of the model(s).\n\nFor Cisco IOS and NXOS devices (which use CLI-based NEDs), the built-in NCS `live-status` action(s) are used to send raw CLI commands to the device.  For example, the `get_mac_address()` method will send a `show mac address-table` CLI command.  For both IOS and NXOS we use [ntc_templates](https://github.com/networktocode/ntc-templates) to parse the raw text output into structured data.\n\nFor Juniper devices, since the NED uses NETCONF for all device communications, we instead call the `<get-ethernet-switching-table-information>` RPC directly.  Since this RPC is modelled in YANG, we can then parse the results directly using the maagic API.\n\nAll the nitty-gritty details of parsing the data retrieved directly from the remote device is handled by the platform-specific model implementation for that device.  Each model normalizes the data using the dataclasses defined in [base.py](./umnet_pyncs/state/models/base.py).  The intention is to makes it simpler for NCS actions/services to use this module, as well as making it easier to develop/maintain.\n\n**NB**: this implementation currently relies on an additional template for NXOS that handles parsing `show ip arp detail vrf all` -- see [PR# 1204](https://github.com/networktocode/ntc-templates/pull/1204).\n\n## Developer testing\n\nInstall the [ipython-superuser](https://github.com/NSO-developer/ipython-superuser) package into your local NSO development environment.  Use the REPL e.g:\n\n``` python\nimport logging\nfrom umnet_pyncs.state import StateManager\n\nlogging.basicConfig(level=logging.DEBUG)\n\ndevices = [d for d in root.devices.device]\n\nm = StateManager()\n\n# NB: m.get_state() expects a list of devices\ninterfaces = m.get_state(devices, ["get-interface-details"])\n    \nprint(interfaces)\n\nm.__exit__()\n```\n',
    'author': 'Nick Grundler',
    'author_email': 'grundler@umich.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import List, Dict, Optional
from ipaddress import IPv4Address

from ntc_templates.parse import parse_output
from netaddr import EUI, mac_unix_expanded

from .base import BaseDevice, ARPTableEntry, MACTableEntry, Interface


ADMIN_STATUS = {
    "connected": "enabled",
    "suspndByV": "enabled",
    "notconnec": "enabled",
    "disabled": "disabled",
    "sfpAbsent": "enabled",
    "xcvrAbsen": "enabled",
    "suspnd": "enabled",
    "suspended": "enabled",
    "down": "disabled",  # Vlan1  --   down      routed    auto    auto
}

OPER_STATUS = {
    "connected": "up",
    "suspndByV": "down",
    "notconnec": "down",
    "disabled": "down",
    "sfpAbsent": "down",
    "xcvrAbsen": "down",
    "suspnd": "down",
    "suspended": "down",
    "down": "down",  # Vlan1  --   down      routed    auto    auto
}

DUPLEX = {
    "half": "half",
    "full": "full",
    "auto": "full",
}

SPEED = {
    "10": "10",
    "100": "100",
    "1000": "1000",
    "auto": "auto",
    "a-10": "10",
    "a-100": "100",
    "a-1000": "1000",
    "a-10G": "10000",
    "100G": "auto",  # Eth1/97  ...     100G      QSFP-100G-LR4
}


class NXOS(BaseDevice):
    def _run_cmd(self, command: str) -> str:
        """
        use NCS live-status exec to issue a raw show command towards a device.
        platform-specific models are expected to parse this output and return
        structured data.

        :param command: string CLI command, minus the 'show', e.g. 'interface status'
        :returns: raw string output from the device
        """
        show = self.device.live_status.__getitem__("exec").show
        inp = show.get_input()
        inp.args = [command]

        return show.request(inp)

    def get_interface_details(self, interface: Optional[str] = None) -> List[Interface]:
        """
        Gathers interface operational data from an NXOS device by parsing 'show interfaces status'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_interface_status

        parsed_sample:
        - port: "mgmt0"
          name: "--"
          status: "connected"
          vlan: "routed"
          duplex: "full"
          speed: "100"
          type: "--"
        """
        # ncs automatically adds the 'show' to the front of the cmd
        self.log.debug(f"sending 'show interface status' to {self.device.name}")
        command = f"interface {interface} status" if interface else "interface status"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

        output = []
        for entry in parsed:
            interface = Interface(
                name=entry["port"],
                description=entry["name"],
                admin_status=ADMIN_STATUS[entry["status"]],
                oper_status=OPER_STATUS[entry["status"]],
                duplex=DUPLEX[entry["duplex"]],
                speed=SPEED[entry["speed"]],
            )

            if interface.name.startswith("Vlan"):
                interface.is_logical = True

            output.append(interface)

        return output

    def get_bfd_neighbors(self, interface: Optional[str] = None) -> List[Dict]:
        """
        Gathers bfd operational data from an NXOS device by parsing 'show bfd neighbors'

        see the ntc_templates test data for this template for details on output structure:
        https://gitlab.umich.edu/grundler/ntc-templates/-/tree/add-nxos-show-bfd-neighbors/tests/cisco_nxos/show_bfd_neighbors

        ---
        parsed_sample:
          - our_addr: "172.23.0.18"
            neigh_addr: "172.23.0.19"
            ld_rd: "1090540255/1090566886"
            rh_rs: "Up"
            holddown: "862(3)"
            state: "Up"
            interface: "Eth1/10"
            vrf: "default"
            type: "SH"
        """
        self.log.debug(f"sending 'show bfd neighbors' to {self.device.name}")
        command = "bfd neighbors"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_ospf_neighbors(self, interface: Optional[str] = None) -> List[Dict]:
        """
        Gathers live ospf operational data from an NXOS device by parsing 'show
        ip ospf neighbor'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_ip_ospf_neighbor

        ---
        parsed_sample:
          - interface: "Vlan999"
            local_ipaddr: "11.11.11.11"
            neighbor_ipaddr: "10.0.0.1"
            ospf_pid: "1111"
            state: "FULL/ -"
            uptime: "8w6d"
            vrf: "CUSTVRF1"
        """
        self.log.debug(f"sending 'show ip ospf neighbor' to {self.device.name}")
        command = "ip ospf neighbor"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_lldp_neighbors(self, interface: Optional[str] = None) -> List[Dict]:
        """
        Gathers live operational data about physically connected devices from an
        NXOS device by parsing 'show lldp neighbors'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_lldp_neighbors

        ---
        parsed_sample:
          - neighbor: "nx-osv9000-3-long-name.com"
            local_interface: "Eth1/1"
            neighbor_interface: "Ethernet1/1"
        """
        self.log.debug(f"sending 'show lldp neighbors' to {self.device.name}")
        command = "lldp neighbors"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_transceiver_details(self, interface: Optional[str] = None) -> List[Dict]:
        """
        Gathers live operational state DOM information from an NXOS device by
        parsing 'show interface tranceiver details'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_interface_transceiver_details

        ---
        parsed_sample:
          - interface: "Ethernet3/1"
            lane: ""
            temperature_value: "35.66"
            temperature_alarm_high: "75.00"
            temperature_alarm_low: "-5.00"
            temperature_warn_high: "70.00"
            temperature_warn_low: "0.00"
            voltage_value: "3.29"
            voltage_alarm_high: "3.63"
            voltage_alarm_low: "2.97"
            voltage_warn_high: "3.46"
            voltage_warn_low: "3.13"
            amps_value: "N/A"
            amps_alarm_high: "11.80"
            amps_alarm_low: "4.00"
            amps_warn_high: "10.80"
            amps_warn_low: "5.00"
            rx_value: "-30.45"
            rx_alarm_high: "2.00"
            rx_alarm_low: "-13.90"
            rx_warn_high: "-1.00"
            rx_warn_low: "-9.90"
            tx_value: "-14.05"
            tx_alarm_high: "1.69"
            tx_alarm_low: "-11.30"
            tx_warn_high: "-1.30"
            tx_warn_low: "-7.30"
        """
        self.log.debug(
            f"sending 'show interface transceiver details' to {self.device.name}"
        )
        command = "interface transceiver details"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_arp_table(
        self, address: Optional[str] = None, vrf: Optional[str] = None
    ) -> List[ARPTableEntry]:
        """
        Gathers live operational state ARP information from an NXOS device by
        parsing 'show ip arp detail vrf all'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_ip_arp_detail

        NB: see https://github.com/networktocode/ntc-templates/pull/1204 -- this PR adds support for
        adding 'vrf all' to the CLI command which is required for our use case.

        ---
        parsed_sample:
          - address: "192.168.56.2"
            age: "00:17:02"
            mac: "5087.89a1.d8d5"
            interface: "Ethernet1/2"
            physical_interface: "Ethernet1/2"
            flags: "+"
            vrf: "PRIMARY"
          - address: "90.10.10.2"
            age: "00:02:55"
            mac: "000d.ece7.df7c"
            interface: "Vlan900"
            physical_interface: "Ethernet1/12"
            flags: ""
            vrf: "PRIMARY"
        """
        self.log.debug(f"sending 'show ip arp detail vrf all' to {self.device.name}")
        command = "ip arp detail vrf all"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_nxos",
            command="show ip arp detail vrf all",
            data=reply.result,
        )

        if address is not None:
            normalized_address = IPv4Address(address)

        output = []
        for entry in parsed:
            if entry["mac"] == "INCOMPLETE":
                continue

            if entry["physical_interface"] == "":
                interface = entry["interface"]
            else:
                interface = entry["physical_interface"]

            entry_vrf = None if entry["vrf"] == "" else entry["vrf"]
            normalized_entry_address = IPv4Address(entry["address"])

            if address is None or normalized_address == normalized_entry_address:
                arp = ARPTableEntry(
                    ip_address=IPv4Address(entry["address"]),
                    mac_address=EUI(entry["mac"], dialect=mac_unix_expanded),
                    interface=interface,
                    vrf=entry_vrf,
                )
                output.append(arp)

        return output

    def get_mac_table(self, mac: Optional[str] = None) -> List[MACTableEntry]:
        """
        Gathers learned MAC addresses from an NXOS device by parsing 'show mac address-table dynamic'.

        :mac: optionally filter device output by MAC address

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_mac_address-table

        ---
        parsed_sample:
          - vlan: "100"
            mac: "5087.89a1.de75"
            type: "static"
            age: "-"
            secure: "F"
            ntfy: "F"
            ports: "sup-eth1(R)"
        """
        command = "mac address-table dynamic"
        if mac is not None:
            normalized_mac = EUI(mac, dialect=mac_unix_expanded)
            command = command + f" address {mac}"

        self.log.debug(f"sending 'show {command}' to {self.device.name}")
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_nxos",
            command="show mac address-table",
            data=reply.result,
        )

        output = []
        for entry in parsed:
            try:
                vid = int(entry["vlan"])
            except ValueError:
                vid = None

            normalized_entry_mac = EUI(entry["mac"], dialect=mac_unix_expanded)

            if mac is None or normalized_mac == normalized_entry_mac:
                mac_entry = MACTableEntry(
                    address=normalized_entry_mac,
                    vlan_id=vid,
                    interface=entry["ports"],
                )
                output.append(mac_entry)

        return output
